#uses "fwGeneral/fwExceptionHandling"
#uses "AlarmScreenNg/AlarmScreenNg.ctl"    // Common definitions for AS NG
#uses "AlarmScreenNg/AlarmScreenNgActions.ctl"    // The names of 'standard' actions in AS EWO
#uses "AlarmScreenNg/classes/AsNgAccessControl.ctl"  // Constants for action states
#uses "AlarmScreenNg/AlarmScreenNgEwo.ctl" // Common processing of AS EWO callbacks
#uses "fwAlarmScreenNg/AlarmScreenJCOP.ctl"  // UNICOS reuses some of JCOP definitions
#uses "AlarmScreenNg/AlarmScreenNgDisplay"  // functions to open 'standard' alarm panels
#uses "unAlarmScreenNg/classes/AsNgAcknowledgeUNICOS.ctl"  // Implementation of acknowledgemtn for UNICOS
#uses "unGenericObject.ctl"
#uses "unGraphicalFrame.ctl"
#uses "unicos_declarations_core.ctl"
#uses "unGenericDpFunctions.ctl"
#uses "unDeviceListUpdate.ctl"             // for g_dsDeviceType
#uses "unGenericFunctions_core"
#uses "unConfigGenericFunctions"

/** @file AlarmScreenUNICOS.ctl
  UNICOS-cpecific support for NextGen Alarm Screen
*/

/** @name UNICOS-specific names of alarm properties, as they appear in alarm data mapping from AS EWO
*/
//@{
const string ALARM_PROP_UN_TIME_UTC = "timeUtc";          ///< UTC alarm time
const string ALARM_PROP_UN_APPLCIATION = "application";   ///< UNICOS application name
const string ALARM_PROP_UN_DEVICE_TYPE = "deviceType";    ///< UNICOS device type name
const string ALARM_PROP_UN_ALIAS = "alias";               ///< DP alias
const string ALARM_PROP_UN_DOMAIN = "domain";             ///< UNICOS device domain name
const string ALARM_PROP_UN_NATURE = "nature";             ///< UNICOS device nature name
const string ALARM_PROP_UN_DESCRIPTION = "description";   ///< UNICOS device description
const string ALARM_PROP_UN_ALARM_TYPE = "alarmType";      ///< UNICOS alarm type
const string ALARM_PROP_UN_MASTER = "master";             ///< UNICOS master device name (for AA and DA)
const string ALARM_PROP_UN_ALARM_LEVEL = "alarmLevel";    ///< UNICOS alarm level (0/1/2/3 = Info/Warning/Alarm/Safety alarm)
const string ALARM_PROP_UN_PANEL = "unPanel";             ///< UNICOS device panel name
const string ALARM_PROP_UN_S = "s";                       ///< UNICOS column titled 'S', see mouseTableEvent

//@}

/** @name Replacement for constances declared in unAlarmScreen.ctl in order to remove dependency on that library */
//@{

/// Replacement for UNALARMSCREEN_FILTER_ALL: the string that appears in filter with meaning 'all'
const string AS_NG_UN_ALARM_FILTER_ALL = "*";

/// Replacement for UNALARMSCREEN_LIST_NAME: the string that appears in selection list of combobox
const string AS_NG_UN_ALARM_LIST = "List...";

/// Replacement for UNALARMSCREEN_POPUPMENU_INDEX_COPY_LINE: the index of popup menu item "Copy to clipboard"
const int AS_NG_UN_MENU_INDEX_COPY_LINE = 100000;

//@}

/// Flag for activating debug messages in this library, one flag for all functions
private const string ASNG_UN_DBG = "ASNG_DBG";

/** @name Public functions */
//@(

/**
 * Get list of comletions for filter on given alarm property
 *
 * @param sPropName The name of alarm property for which filter is configured
 * @param dsSelectedSystems The list of selected system names
 * @return mapping with 3 fields (keys are strings):
 *    - "type" The type of completions used for this filter, see enum FilterCompletionsType
 *    - "list" The list of completion strings for this filter
 *    - "map"  The list of completion strings for this filter in case complitions are related
 *            to connected systems. Key is string (system name), value is dyn_string = list of
 *            completions if this system is connected
 */
mapping AlarmScreenUNICOS_getFilterCompletions(string sPropName, const dyn_string &dsSelectedSystems)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(" + sPropName + ")", dsSelectedSystems);
  // Default is no completions
  mapping mResult;
  mResult[AS_FILTER_COMPLETION_KEY_TYPE] = (int)FilterCompletionsType::None;
  // Find completions depending on alarm property name
  switch(sPropName)
  {
    case ALARM_PROP_UN_APPLCIATION:
      _AlarmScreenUNICOS_getFilterCompletions_application(dsSelectedSystems, mResult);
      break;
    case ALARM_PROP_UN_NATURE:
      _AlarmScreenUNICOS_getFilterCompletions_nature(dsSelectedSystems, mResult);
      break;
    case ALARM_PROP_UN_DEVICE_TYPE:
      _AlarmScreenUNICOS_getFilterCompletions_deviceType(dsSelectedSystems, mResult);
      break;
    case ALARM_PROP_UN_DOMAIN:
      _AlarmScreenUNICOS_getFilterCompletions_domain(dsSelectedSystems, mResult);
      break;
    /*
    case ALARM_PROP_ACK_STATE:  // TODO: REMOVE AFTER TESTS
      mResult[AS_FILTER_COMPLETION_KEY_TYPE] = (int)FilterCompletionsType::Dynamic;
      mResult[AS_FILTER_COMPLETION_KEY_LIST] = makeDynString("Direction");
      break;
    */
  }
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(" + sPropName + "): result", mResult);
  return mResult;
}

/**
 * Get list of comletions for slave filter on given alarm property
 *
 * @param sPropName The name of alarm property for which filter is configured
 * @param mFilters Current settings for all 'master' filters of this slave:
 *        - key: alarm property name of filter
 *        - value: current filter value for this alarm property
 * @return mapping with 1 field (key is string):
 *    - "list" The list of completion strings for this filter
 */
mapping AlarmScreenUNICOS_getSlaveFilterCompletions(string sPropName, const mapping &mFilters)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(" + sPropName + ")", mFilters);
  // Combine the result
  mapping mResult;
  mResult[AS_FILTER_COMPLETION_KEY_LIST] = makeDynString();
  /*
  if(sPropName == ALARM_PROP_ACK_STATE) // TODO: REMOVE AFTER TESTS
  {
    string sMaster = mFilters[ALARM_PROP_DIRECTION];
    switch(sMaster)
    {
    case "CAME":
      mResult[AS_FILTER_COMPLETION_KEY_LIST] = makeDynString("CAME #1", "CAME #2", "Common CAME+WENT", "CAME #3", "CAME #4");
      break;
    case "WENT":
      mResult[AS_FILTER_COMPLETION_KEY_LIST] = makeDynString("WENT #1", "WENT #2", "WENT #3", "WENT #4", "Common CAME+WENT", "WENT #5", "WENT #6");
      break;
    }
  }
  */
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(" + sPropName + "): result", mResult);
  return mResult;
}

/**
 * Process mouse event in alarm table cell. TODO - code and description
 * The variant of calling this from C++ code doesn't work for UNICOS, postponed...
 */
void AlarmScreenUNICOS_mouseTableEvent(const mapping &mArgs)
{
  const mapping mSource = mArgs["mSource"];
  const mapping mAlarmData = mArgs["mAlarm"];
  const int iMode = mArgs["sourceMode"];
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "():", mSource, mAlarmData, iMode);
  _AlarmScreenUNICOS_initHMI();

  if((mSource["type"] == 0)  // Mouse pressed event
    && (mSource["button"] == 2))  // Right mouse button
  {
    dyn_string dsMenu = _AlarmScreenUNICOS_buildMenu(mSource, mAlarmData, iMode);
    if(dynlen(dsMenu) > 0)
    {
      int iAnswer;
      popupMenuXY(dsMenu, mSource["screenX"], mSource["screenY"], iAnswer);
      if(iAnswer > 0)
      {
        AlarmScreenUNICOS_treatRightClickAnswer(mSource, mAlarmData, dsMenu, iAnswer);
      }
    }
  }
  else if((mSource["type"] == 1)  // Mouse click event
    && (mSource["button"] == 1)  // Left mouse button
    && (mSource["columnSource"] == 0))
  {
    string sDpe = mAlarmData[ALARM_PROP_FULL_DPE];
    if(!dpExists(sDpe))
    {
      return;
    }
    switch(mSource["columnId"])
    {
    case ALARM_PROP_ACK_STATE_SUMMARY:  // Acknowledge when clicked on 'Ack.' column, only in onlinemode
      if((iMode == AS_NG_EWO_MODE_ONLINE) && (mAlarmData[ALARM_PROP_ACK_STATE_SUMMARY] == ALARM_ACK_STATE_SUMMARY_ACKABLE_OLDEST))
      {
        dyn_mapping dmAlarms = makeDynMapping(mAlarmData);
        _AlarmScreenUNICOS_ackMultiple(dmAlarms);
      }
      break;
    case ALARM_PROP_UN_S:  // Open default device panel
      if((iMode == AS_NG_EWO_MODE_ONLINE) && (mappingHasKey(mAlarmData, ALARM_PROP_UN_S)))
      {
        if(mAlarmData[ALARM_PROP_UN_S] == "S")  // magic char indicating existence of default panel
        {
          _AlarmScreenUNICOS_defaultPanel(unGenericDpFunctions_getDpName(mAlarmData[ALARM_PROP_FULL_DPE]), mAlarmData[ALARM_PROP_UN_DESCRIPTION]);
        }
      }
      break;
    }
  }
}

/**
 * Process ctrlActionRequested() event of AS EWO in UNICOS-specific way
 * @param mArgs event arguments, see function AlarmScreenNgEwo_ctrlActionRequested()
 */
void AlarmScreenUNICOS_ctrlActionRequested(const mapping &mArgs)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "():", mArgs);
  _AlarmScreenUNICOS_initHMI();
  shape ewo = mArgs["ewoShape"];
  if(ewo == 0)
  {
    throwError(makeError("", PRIO_SEVERE, ERR_IMPL, 76,  // 00076,Invalid argument in function
                         __FUNCTION__ + "(): EWO shape not found in arguments"));
    return;
  }
  dyn_mapping dmAlarms;
  const int iMode = mArgs["sourceMode"];
  switch(mArgs["sAction"])
  {
  case AS_EWO_ACTION_BASIC_CONFIG_APPLIED:
    _AlarmScreenUNICOS_basicConfigApplied(mArgs);
    break;
  case AS_EWO_ACTION_ACK_SELECTED:
    if(iMode == AS_NG_EWO_MODE_ONLINE)
    {
      dmAlarms = ewo.getForAcknowledgementSelected();
      _AlarmScreenUNICOS_ackMultiple(dmAlarms);
    }
    break;
  case AS_EWO_ACTION_ACK_GONE:
    if(iMode == AS_NG_EWO_MODE_ONLINE)
    {
      dmAlarms = ewo.getForAcknowledgementAllGone();
      _AlarmScreenUNICOS_ackMultiple(dmAlarms);
    }
    break;
  case AS_EWO_ACTION_ACK_ALL:
    if(iMode == AS_NG_EWO_MODE_ONLINE)
    {
      dmAlarms = ewo.getForAcknowledgementAll();
      _AlarmScreenUNICOS_ackMultiple(dmAlarms);
    }
    break;
  default:
    AlarmScreenNgActionProcessing_process(mArgs);  // Default processing
  }
}

/**
 * Process mouse event in alarm table cell. TODO - code and description
 * The variant of calling this from C++ code doesn't work for UNICOS, postponed...
 * @note This function is not used, remains here as first variant for tests
 */
mapping AlarmScreenUNICOS_mouseTableEvent_broken(const mapping &mSource, const mapping &mAlarmData)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "():", mSource, mAlarmData);
  mapping mResult;
  if((mSource["type"] == 0)  // Mouse pressed event
    && (mSource["button"] == 2))  // Right mouse button
  {
    dyn_string dsMenu = _AlarmScreenUNICOS_buildMenu(mSource, mAlarmData);
    if(dynlen(dsMenu) > 0)
    {
      int iAnswer;
      popupMenuXY(dsMenu, mSource["screenX"], mSource["screenY"], iAnswer);
      if(iAnswer > 0)
      {
        AlarmScreenUNICOS_treatRightClickAnswer(mSource, mAlarmData, dsMenu, iAnswer);
      }
    }
  }
  return mResult;
}

/**
 * Process user selection from popup menu on alarm table. The whole logic of processing
 * follows the logic of unAlarmScreen_treatRightClickAnswer(), as much as possible.
 * @param mSource Parameters of event source (what row/column in table was clicked,
 *        mouse pointer position etc.)
 * @param mAlarmData Properties of all alarms, selected in the table
 * @param dsMenu The description of popup menu that was hown to user
 * @param iAnswer The ID of menu item, selected by user
 * @return TODO
 */
mapping AlarmScreenUNICOS_treatRightClickAnswer(const mapping &mSource, const mapping &mAlarmData, const dyn_string &dsMenu, int iAnswer)
{
  if(!_AlarmScreenUNICOS_menuAlarmSpecific(mSource, mAlarmData, iAnswer))
  {
    _AlarmScreenUNICOS_menuCpcGenericAction(mSource, mAlarmData, dsMenu, iAnswer);
  }
}

/**
 * Copy of function unAlertPanel_getCleanModuleName() from unAlertPanel component.
 *
 * FWAH-364: Make sure that the name does not contain any fancy character when assigning
 * a name to a module.
 * @param sOrigName Original name to be cleared of fancy characters
 * @return clear name to be used as module name
 */
string AlarmScreenUNICOS_getCleanModuleName(string sOrigName)
{
  string sCleanName = sOrigName;

  strreplace(sCleanName, ".", "_dot_");
  strreplace(sCleanName, "/", "_slash_");
  strreplace(sCleanName, "\\", "_bslash_");
  strreplace(sCleanName, ":", "_semicol_");
  strreplace(sCleanName, "*", "_star_");
  strreplace(sCleanName, "#", "_pound_");
  strreplace(sCleanName, ",", "_comma_");
  strreplace(sCleanName, ";", "_coldot_");

  return sCleanName;
}

/**
 * Setup a default Application filter with the corresponding values if a Filtered Application has been configured
 */
void AlarmScreenUNICOS_setFilteredApplication()
{
  string applicationList = strjoin(g_unicosHMI_dsApplicationList, ";") + ";";
  string filterJsonTemplate = "{\"filterRulesSets\":[{\"filters\":[{\"filter\":{\"pattern\":\"%s\",\"source\":\"this\",\"valueId\":\"application\"}}],\"include\":true}],\"matchCase\":false}";
  string filterJson;
  sprintf(filterJson, filterJsonTemplate, applicationList);
  AsEwo.applyFilter(filterJson);
}

//@}

/** @name Private functions: initialization of some globals for UNICOS UI */
//@{

/**
 * Simplified version of unGraphicalFrame_initHMI(), the main differences are:
 *   - added check if initialization was already done
 *   - not all of initializations are performed
 */
private void _AlarmScreenUNICOS_initHMI()
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(): START");
  if((dynlen(g_dsPrivilegeAccessControl) > 0)
    && (dynlen(g_dsDomainAccessControl) > 0))  // Use as indication of 'initialization was done'
  {
    DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(): FINISH, already initialized");
    return;
  }
  // initialize the g_bUnGraphicalFrame_RDB
  dpGet("_DataManager.UseRDBArchive", g_bUnGraphicalFrame_RDB);
  // get the list of installed device type
  unGenericDpFunctions_getUnicosObjects(g_dsDeviceType);

  dpConnect("unGraphicalFrame_domainCB", "_Areas.UserName");// TODO: is this needed ?
  g_dsPrivilegeAccessControl = makeDynString("monitor", "operator", "expert", "admin");

  // setup call back for the user panel.
  dyn_string exceptionInfo;  // TODO: how to process this exceptionInfo ?
  unConfigGenericFunctions_createDp(unGraphicalFrame_getInternalTriggerName(), GRAPHICALFRAME_DPTYPE, exceptionInfo);
  unGraphicalFrame_initUserPanel();

  // load the device type config in memory
  unGenericObject_InitializeObjects(exceptionInfo);

  // loat in memory the list of front-end device type
  // TODO: the variable is declared in UNICOS panel, do we need this initialization? unGenericDpFunctions_getFrontEndDeviceType(g_dsConfiguredFrontEndType);
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(): FINISH, new initialization");
}

//@}

/**
 * Process ctrlActionRequested() event of AS EWO in UNICOS-specific way; the case when event
 * is AS_EWO_ACTION_BASIC_CONFIG_APPLIED. The main goal of this method is performing UNICOS-specific
 * initialization of AS EWO
 * @param mArgs event arguments, see function AlarmScreenNgEwo_ctrlActionRequested()
 */
private void _AlarmScreenUNICOS_basicConfigApplied(const mapping &mArgs)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "():", mArgs);

  // First perform generic initialization...
  AlarmScreenNgActionProcessing_process(mArgs);  // Default processing

  // ... then specific one. Hide popup menu configuration that doesn't make sense for UNICOS
  shape ewo = mArgs["ewoShape"];
  if(ewo == 0)
  {
    throwError(makeError("", PRIO_SEVERE, ERR_IMPL, 76,  // 00076,Invalid argument in function
                         __FUNCTION__ + "(): EWO shape not found in arguments"));
    return;
  }
  ewo.setFixedActionState(AS_EWO_ACTION_ADMIN_POPUP_MENU, ALARM_SCREEN_ACCESS_ACTION_HIDE);

  // Read and apply time zone settings for UNICOS application
  bool bUseUTC = false;
  if(dpExists(UN_APPLICATION_DPNAME)) {
    dpGet(UN_APPLICATION_DPNAME + ".UTC", bUseUTC);
    dyn_errClass deErrors = getLastError();
    if(!deErrors.isEmpty()) {
      throwError(deErrors);
    }
  }
  ewo.timeSpec = bUseUTC ? AS_NG_EWO_TIMESPEC_UTC : AS_NG_EWO_TIMESPEC_LOCAL;
}

/** @name Private functions: processing of filter completions */
//@(

/**
 * Get completions for filter on application name
 * @param dsSelectedSystems The list of selected system names
 * @param mResult mapping where result shall be placed, 3 fields are available (keys are strings):
 *    - "type" The type of completions used for this filter, see enum FilterCompletionsType
 *    - "list" The list of completion strings for this filter
 *    - "map"  The list of completion strings for this filter in case complitions are related
 *            to connected systems. Key is string (system name), value is dyn_string = list of
 *            completions if this system is connected
 */
private void _AlarmScreenUNICOS_getFilterCompletions_application(const dyn_string &dsSelectedSystems, mapping &mResult)
{
  mResult[AS_FILTER_COMPLETION_KEY_TYPE] = (int)FilterCompletionsType::SysRelated;
  mapping mSysData;
  for(int iSysIdx = dynlen(dsSelectedSystems) ; iSysIdx > 0 ; iSysIdx--)
  {
    mSysData[dsSelectedSystems[iSysIdx]] = _AlarmScreenUNICOS_getSystemApplications(dsSelectedSystems[iSysIdx]);
  }
  mResult[AS_FILTER_COMPLETION_KEY_MAP] = mSysData;
}

/**
 * Get list of applications for given system name. This is a bit simplified version of function
 * _unTreeDeviceOverview_getFrontEndNameApplicationFromFile()
 * @param sSystemName The name of system
 * @param sFileNameSuffix Suffix for file name with information
 * @return list of application names for this system
 */
private dyn_string _AlarmScreenUNICOS_getSystemApplications(string sSystemName, string sFileNameSuffix="_unTreeDeviceOverview_FECharacteristics.dat")
{
  dyn_string dsResult;

  strreplace(sSystemName, ":", "");
  string sFileName = getPath(DATA_REL_PATH) + "_" + sSystemName + sFileNameSuffix;
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(): reading file: " + sFileName);
  if(access(sFileName, F_OK) != 0)
  {
    DebugN(__FUNCTION__ + "(): file does not exist: " + sFileName);
    return dsResult;
  }
  if(access(sFileName, R_OK) != 0)
  {
    DebugN(__FUNCTION__ + "(): file can't be read: " + sFileName);
    return dsResult;
  }

  string sBuffer;
  fileToString(sFileName, sBuffer);
  dyn_string dsSplit = strsplit(sBuffer, "\n");
  for(int iLineIdx = dynlen(dsSplit) ; iLineIdx > 0 ; iLineIdx--)
  {
    dyn_string dsTemp = strsplit(dsSplit[iLineIdx], ";");
    // skip first 3 names: 1 = FrontEnd, 2 = FrontEndDp, 3 = FrontEndDpType
    for(int idx = dynlen(dsTemp) ; idx > 3 ; idx--)
    {
      dynAppend(dsResult, dsTemp[idx]);
    }
  }

  // Load new Application filter values to the autocompletion mechanism in case of running in a Filtered Application
  if (g_unicosHMI_bApplicationFilterInUse)
  {
    synchronized(g_unicosHMI_dsApplicationList)
    {
      dsResult = dynIntersect(dsResult, g_unicosHMI_dsApplicationList);
    }
  }

  return dsResult;
}

/**
 * Get completions for filter on device type
 * @param dsSelectedSystems The list of selected system names
 * @param mResult mapping where result shall be placed, 3 fields are available (keys are strings):
 *    - "type" The type of completions used for this filter, see enum FilterCompletionsType
 *    - "list" The list of completion strings for this filter
 *    - "map"  The list of completion strings for this filter in case complitions are related
 *            to connected systems. Key is string (system name), value is dyn_string = list of
 *            completions if this system is connected
 */
private void _AlarmScreenUNICOS_getFilterCompletions_deviceType(const dyn_string &dsSelectedSystems, mapping &mResult)
{
  mResult[AS_FILTER_COMPLETION_KEY_TYPE] = (int)FilterCompletionsType::Static;
  dyn_string dsCompletionList;
  unGenericDpFunctions_getUnicosObjects(dsCompletionList);
  dynUnique(dsCompletionList);
  mResult[AS_FILTER_COMPLETION_KEY_LIST] = dsCompletionList;
}

/**
 * Get completions for filter on device domain
 * @param dsSelectedSystems The list of selected system names
 * @param mResult mapping where result shall be placed, 3 fields are available (keys are strings):
 *    - "type" The type of completions used for this filter, see enum FilterCompletionsType
 *    - "list" The list of completion strings for this filter
 *    - "map"  The list of completion strings for this filter in case complitions are related
 *            to connected systems. Key is string (system name), value is dyn_string = list of
 *            completions if this system is connected
 */
private void _AlarmScreenUNICOS_getFilterCompletions_domain(const dyn_string &dsSelectedSystems, mapping &mResult)
{
  mResult[AS_FILTER_COMPLETION_KEY_TYPE] = (int)FilterCompletionsType::SysRelated;
  mapping mSysData;
  for(int iSysIdx = dynlen(dsSelectedSystems) ; iSysIdx > 0 ; iSysIdx--)
  {
    dyn_string dsDomains;
    dynClear(dsDomains);
    _AlarmScreenUNICOS_getSubSystemFromFile(dsSelectedSystems[iSysIdx], 1, dsDomains);
    int iAllPos = dynContains(dsDomains, AS_NG_UN_ALARM_FILTER_ALL);
    if(iAllPos > 0)
    {
      dynRemove(dsDomains, iAllPos);
    }
    int iListPos = dynContains(dsDomains, AS_NG_UN_ALARM_LIST);
    if(iListPos > 0)
    {
      dynRemove(dsDomains, iListPos);
    }
    mSysData[dsSelectedSystems[iSysIdx]] = dsDomains;
  }
  mResult[AS_FILTER_COMPLETION_KEY_MAP] = mSysData;
}

/**
 * Get completions for filter on device nature
 * @param dsSelectedSystems The list of selected system names
 * @param mResult mapping where result shall be placed, 3 fields are available (keys are strings):
 *    - "type" The type of completions used for this filter, see enum FilterCompletionsType
 *    - "list" The list of completion strings for this filter
 *    - "map"  The list of completion strings for this filter in case complitions are related
 *            to connected systems. Key is string (system name), value is dyn_string = list of
 *            completions if this system is connected
 */
private void _AlarmScreenUNICOS_getFilterCompletions_nature(const dyn_string &dsSelectedSystems, mapping &mResult)
{
  mResult[AS_FILTER_COMPLETION_KEY_TYPE] = (int)FilterCompletionsType::SysRelated;
  mapping mSysData;
  for(int iSysIdx = dynlen(dsSelectedSystems) ; iSysIdx > 0 ; iSysIdx--)
  {
    dyn_string dsNatures;
    dynClear(dsNatures);
    _AlarmScreenUNICOS_getSubSystemFromFile(dsSelectedSystems[iSysIdx], 2, dsNatures);
    int iAllPos = dynContains(dsNatures, AS_NG_UN_ALARM_FILTER_ALL);
    if(iAllPos > 0)
    {
      dynRemove(dsNatures, iAllPos);
    }
    int iListPos = dynContains(dsNatures, AS_NG_UN_ALARM_LIST);
    if(iListPos > 0)
    {
      dynRemove(dsNatures, iListPos);
    }
    mSysData[dsSelectedSystems[iSysIdx]] = dsNatures;
  }
  mResult[AS_FILTER_COMPLETION_KEY_MAP] = mSysData;
}

/**
 * Read from configuration file
 * This is a bit simplified version of original function _unTreeDeviceOverview_getSubSystemFromFile()
 * @param sSystemName System name being processed
 * @param iSubSysIndex Which of 'subsystems' is required (1 = doman, 2 = nature)
 * @param dsResult Variable where resulting names will be placed
 * @param sFileNameSuffix Constant part of file name
 */
private void _AlarmScreenUNICOS_getSubSystemFromFile(string sSystemName, int iSubSysIndex, dyn_string &dsResult, string sFileNameSuffix="_unTreeDeviceOverview_subSystem.dat")
{
  strreplace(sSystemName, ":", "");
  string sFileName = getPath(DATA_REL_PATH) + "_" + sSystemName + sFileNameSuffix;

  string sBuffer;
  fileToString(sFileName, sBuffer);
  dyn_string dsSplit = strsplit(sBuffer, "\n");
  if((dynlen(dsSplit) == 2) && (1 <= iSubSysIndex) && (iSubSysIndex <= 2)) {
    dsResult = strsplit(dsSplit[iSubSysIndex], ";");
  }
}

//@}


/** @name Private functions: preparing popup menu */
//@(

/**
 * Build popup menu for right mouse button event in alarm table. The functionality
 * is copied from unAlarmScreen_treatRightClickAnswer() function, with adaptation for
 * new alarm screen.
 * @param mSource Parameters of event in the table
 * @param mAlarmData All properties of selected alarm(s) in table
 * @param iMode The mode of data source for AS EWO, one of constants AS_NG_EWO_MODE_xxx
 * @return Popup menu description
 */
private dyn_string _AlarmScreenUNICOS_buildMenu(const mapping &mSource, const mapping &mAlarmData, int iMode)
{
  // This never happens because AES_FORCE_CLICK is true if((iRunMode != AES_RUNMODE_RUNNING) && !AES_FORCE_CLICK) { return makeDynString(); }
  dyn_string dsMenuList;
  string sAlias, sDpType;

  string sDpe = mAlarmData[ALARM_PROP_FULL_DPE];
  bool bDpeExists = dpExists(sDpe);
  if(bDpeExists)
  {
    dsMenuList = _AlarmScreenUNICOS_buildBaseDpeMenu(sDpe, sAlias, sDpType);
    _AlarmScreenUNICOS_extendBaseDpeMenu(sAlias, dsMenuList, iMode);
    _AlarmScreenUNICOS_enableDisableAckMenu(mSource, mAlarmData, dsMenuList, iMode);
  }
  else
  {
    dsMenuList = makeDynString("PUSH_BUTTON, DPE deleted, 0, 0",
                               "SEPARATOR",
                               "PUSH_BUTTON, Copy line to clipboard, " + AS_NG_UN_MENU_INDEX_COPY_LINE +", 1",
                               "SEPARATOR");
  }
  dyn_string dsColumnsMenu = _AlarmScreenUNICOS_buildColumnSubMenu(mSource);
  _AlarmScreenUNICOS_buildFinalMenu(dsMenuList, dsColumnsMenu, iMode, bDpeExists);
  return dsMenuList;
}

private dyn_string _AlarmScreenUNICOS_buildBaseDpeMenu(string sDpe, string &sAlias, string &sDpType)
{
  // This part is taken from unAlarmScreen_displayMenu()
  string sDpName = unGenericDpFunctions_getDpName(sDpe);  // Get DpName
  sAlias = unGenericDpFunctions_getAlias(sDpName); // Get alias
  sDpType = dpTypeName(sDpName);
  string sFunction;

  dyn_string dsMenuList;
  dyn_string dsFunctions;
  dyn_string dsAccessOk;
  dyn_string dsExceptions;

  // Alarm panel is the classic UNICOS alarm list.: if (iUnAlertPanel == UNALARMSCREEN_PANEL_ALL_ALARMS)
  // {
  if (sAlias != "" && sDpType != "")
  {
    unGenericButtonFunctionsHMI_isAccessAllowedMultiple(sDpName, sDpType, dsAccessOk, dsExceptions);

    // --------------------------------------
    // --- 1) Build menu
    // --------------------------------------
    unGenericObject_GetFunctions(sDpType, dsFunctions, dsExceptions);
    if (dynlen(dsFunctions) >= UN_GENERICOBJECT_FUNCTION_MENUCONFIG)
    {
      sFunction = dsFunctions[UN_GENERICOBJECT_FUNCTION_MENUCONFIG];
    }
  }

  if (sFunction != "" && sDpName !="")
  {
    if (substr(sDpName, strlen(sDpName)-1, strlen(sDpName))==".")
    {
      sDpName = substr(sDpName, 0, strlen(sDpName) - 1);
    }
    evalScript(
        dsMenuList,
        "dyn_string main(string device, string sType, dyn_string dsAccessOk)" +
        "{" +
        "  dyn_string dsMenuTemp;" +
        "  " + sFunction + "(device, sType, dsAccessOk, dsMenuTemp);" +
        "  return dsMenuTemp;" +
        "}",
        makeDynString(),
        sDpName,
        sDpType,
        dsAccessOk
      );
  }
  /*
  DebugTN(__FUNCTION__ + "():", sDpe, sDpName, sAlias, sDpType, sFunction);
  DebugTN(__FUNCTION__ + "(): dsAccessOk", dsAccessOk);
  DebugTN(__FUNCTION__ + "(): dsMenuList", dsMenuList);
  DebugTN(__FUNCTION__ + "(): dsExceptions", dsExceptions);
  */
  return dsMenuList;
}

/**
 * The basic popup menu was built (or not) for selected DP, this function adds to
 * that menu description items which are always there for any device in alarm screen.
 * @param sAlias Alias for DP in activated row of alarm table
 * @param dsMenuList Previously built description of popup menu, can be empty
 * @param iMode The mode of data source for AS EWO, one of constants AS_NG_EWO_MODE_xxx
 */
private void _AlarmScreenUNICOS_extendBaseDpeMenu(string sAlias, dyn_string &dsMenuList, int iMode)
{
  // This part is taken from unAlarmScreen_displayMenu()
  // Copy line to clipboard
  dynInsertAt(dsMenuList, "SEPARATOR", 1);
  dynInsertAt(dsMenuList, "PUSH_BUTTON, Copy line to clipboard, " + AS_NG_UN_MENU_INDEX_COPY_LINE +", 1", 1);

  if(sAlias != "")
  {
    dynInsertAt(dsMenuList, "PUSH_BUTTON, " + sAlias + ", 1000, 1", 1);
  }

  // Alarm panel is the classic UNICOS alarm list.: if (iUnAlertPanel == UNALARMSCREEN_PANEL_ALL_ALARMS)
  //{
    dynInsertAt(dsMenuList, "SEPARATOR", 2);
  //}
}

/**
 * Build items of submenu for showing/hiding table columns
 * @param mSource Parameters of event in the table
 & @return items for column visibility submenu
 */
private dyn_string _AlarmScreenUNICOS_buildColumnSubMenu(const mapping &mSource)
{
  dyn_string dsMenu;
  dyn_mapping dmColumns = this.getTableColumns();  // The call is originated from AS EWO
  int iColumnCount = dynlen(dmColumns);
  for(int idx = 1 ; idx <= iColumnCount ; idx++)
  {
    if(!dmColumns[idx][AS_TABLE_COL_HIDEABLE])
    {
      continue;
    }
    // CHECK_BUTTON, Name, Return value, Enabled, Checked
    string sItem = "CHECK_BUTTON," + dmColumns[idx][AS_TABLE_COL_TITLE] + "," + (UN_NUM_AESVIEW + idx) + ",1," +
                   (dmColumns[idx][AS_TABLE_COL_VISIBLE] ? "1" : "0");
    dynAppend(dsMenu, sItem);
  }
  if(dynlen(dsMenu) > 0)
  {
    dynAppend(dsMenu, "SEPARATOR");
    dynAppend(dsMenu, "PUSH_BUTTON,View All, " + (UN_NUM_AESVIEW + 100) + ", 1");
  }
  return dsMenu;
}

/**
 * Build the ultimate content of pupup menu.
 * @param dsMenuList Preliminary popup menu description
 * @param dsColumnsMenu Items of 'View' submenu (show/hide table columns)
 * @param iMode The mode of data source for AS EWO, one of constants AS_NG_EWO_MODE_xxx
 * @param bDpeExists <c>true</c> if menu is being built for existing DPE
 */
private void _AlarmScreenUNICOS_buildFinalMenu(dyn_string &dsMenuList, const dyn_string &dsColumnsMenu, int iMode, bool bDpeExists)
{
  // The logic == the last part of unAlarmScreen_viewRightClick() function
  // Find the very first cascade button in existing menu description
  int iLen = dynlen(dsMenuList);
  int iPos = iLen;
  for(int i = 1; i <= iLen && iPos == iLen ; i++)
  {
    if(patternMatch("CASCADE_BUTTON*", dsMenuList[i]))
    {
      iPos = i;
    }
  }

  // Insert AES items
  dynInsertAt(dsMenuList, "CASCADE_BUTTON, View, 1", iPos+1);
  dynInsertAt(dsMenuList, "SEPARATOR", iPos+1);

  // TODO: in original code (unAlarmScreen_viewRightClick) the last part of string is 0 for both items:   // IS-669: problem in case of distributed
  if(bDpeExists)
  {
    string sEnable = (iMode == AS_NG_EWO_MODE_ONLINE) ? "1" : "0";
    dynInsertAt(dsMenuList, "PUSH_BUTTON, Add Comment, " + (UN_NUM_AESVIEW + 300) + ", " + sEnable, iPos + 1);  // UN_NUM_AESVIEW is in unicos_declarations.ctl
    dynInsertAt(dsMenuList, "PUSH_BUTTON, Details, " + (UN_NUM_AESVIEW + 200) + ", " + sEnable, iPos + 1);
    dynInsertAt(dsMenuList, "SEPARATOR", iPos + 1);
  }

  // Column visibility submenu
  dynAppend(dsMenuList, "View");

  iLen = dynlen(dsColumnsMenu);
  for(int i = 1 ; i <= iLen ; i++)
  {
    dynAppend(dsMenuList, dsColumnsMenu[i]);
  }
}

/**
 * Enable/disable menu items of popup menu, related to alarm acknowledgment/masking
 * TODO: the first part of logic is not clear at all, see comments inside this functions
 * @param mSource Parameters of event in the table
 * @param mAlarmData All properties of selected alarm(s) in table
 * @param dsMenuList (partially) prepared description of popup menu
 * @param iMode The mode of data source for AS EWO, one of constants AS_NG_EWO_MODE_xxx
 */
private void _AlarmScreenUNICOS_enableDisableAckMenu(const mapping &mSource, const mapping &mAlarmData, dyn_string &dsMenuList, int iMode)
{
  // The acknowledgement is only allowed in online mode
  bool bAck = (iMode == AS_NG_EWO_MODE_ONLINE);
  if(!bAck)
  {
    // TODO: disable menu item using hardcoded item name... alternative?
    _AlarmScreenUNICOS_disableMenuItem(dsMenuList, "PUSH_BUTTON", "Ack. Alarm");
    _AlarmScreenUNICOS_disableMenuItem(dsMenuList, "PUSH_BUTTON", "Mask PVSS Alarm");
  }
}

/**
 * Disable item in popup menu description, i.e. the very last part shall be ",0" instead of ",1"
 * @param dsMenuList Description of popup menu
 * @param sItemType The type of item to be disabled, typically "PUSH_BUTTON"
 * @param PUSH_BUTTON The label of item to be disabled
 */
private void _AlarmScreenUNICOS_disableMenuItem(dyn_string &dsMenuList, const string &sItemType, const string &sItemLabel)
{
  string sPattern = sItemType + ",*" + sItemLabel + "*";
  for(int idx = dynlen(dsMenuList) ; idx > 0 ; idx--)
  {
    if(patternMatch(sPattern, dsMenuList[idx]))
    {
      dyn_string dsSplit = strsplit(dsMenuList[idx], ",");
      if(dynlen(dsSplit) >= 4)
      {
        dsMenuList[idx] = dsSplit[1] + "," + dsSplit[2] + "," + dsSplit[3] + ",0";
        break; // Exit for loop
      }
    }
  }
}

//@}


/** @name Private functions: processing of user selection from popup menu */
//@(

/**
 * Process user selection from popup menu on alarm table. It was checked before calling this function
 * that user selection is not one of alarm-specific menu items. Thus, selection must be most
 * probably one of items for CPC device.
 * The logic of this function corresponds to unAlarmScreen_treatRightClickAnswer().
 * @param mSource Parameters of event source (what row/column in table was clicked,
 *        mouse pointer position etc.)
 * @param mAlarmData Properties of all alarms, selected in the table
 * @param dsMenu The description of popup menu that was hown to user
 * @param iAnswer The ID of menu item, selected by user
 */
private void _AlarmScreenUNICOS_menuCpcGenericAction(const mapping &mSource, const mapping &mAlarmData, const dyn_string &dsMenu, int iAnswer)
{
  string sDpe = mAlarmData[ALARM_PROP_FULL_DPE];

  string sDpName = unGenericDpFunctions_getDpName(sDpe);  // Get DpName
  string sAlias = unGenericDpFunctions_getAlias(sDpName); // Get alias
  string sDpType = dpTypeName(sDpName);

  dyn_string dsFunctions;
  dyn_string dsExceptions;
  unGenericObject_GetFunctions(sDpType, dsFunctions, dsExceptions);

  string sFunction;
  if(dynlen(dsFunctions) > 0)
  {
    sFunction = dsFunctions[UN_GENERICOBJECT_FUNCTION_HANDLEMENUANSWER];
  }

  if((sFunction != "") && (sDpName !=""))
  {
    if(isFunctionDefined(sFunction))
    {
      dyn_string dsExceptionsBis;
      evalScript(
        dsExceptionsBis,
        "dyn_string main(string deviceName, string sDpType, dyn_string dsMenuList, int menuAnswer)" +
        "{" +
        "  dyn_string dsExceptions;" +
        "  " + sFunction + "(deviceName, sDpType, dsMenuList, menuAnswer);" +
        "  return makeDynString();" +
        "}",
        makeDynString(),
        sDpName,
        sDpType,
        dsMenu,
        iAnswer
      );

      dynAppend(dsExceptions, dsExceptionsBis);
    }
  }

  if(dynlen(dsExceptions) > 0)
  {
    unSendMessage_toExpertException("Right Click", dsExceptions);
  }
}

/**
 * Process menu selection corresponding to one of alarm-specific menu items (in contrast
 * to general popup menu for UNICOS devices).
 * @param mSource Parameters of event in the table
 * @param mAlarmData All properties of selected alarm(s) in table
 * @param dsMenu iAnswer The ID of item selected from menu
 * @return true if selection in menu corresponds to one of alarm-specific menu items, i.e.
 *        if user selection was processed; false otherwise
 */
private bool _AlarmScreenUNICOS_menuAlarmSpecific(const mapping &mSource, const mapping &mAlarmData, int iAnswer)
{
  if(AS_NG_UN_MENU_INDEX_COPY_LINE == iAnswer)
  {
    return _AlarmScreenUNICOS_menuLineToClipboard(mSource, mAlarmData);
  }
  else if(iAnswer == (UN_NUM_AESVIEW + 100)) // View all columns in table
  {
    return _AlarmScreenUNICOS_menuShowAllColumns(mSource);
  }
  else if((iAnswer > UN_NUM_AESVIEW) && (iAnswer < (UN_NUM_AESVIEW + 100))) // Show/hide columns one by one
  {
    return _AlarmScreenUNICOS_menuShowHideColumn(mSource, iAnswer - UN_NUM_AESVIEW);
  }
  else if(iAnswer == (UN_NUM_AESVIEW + 200))
  {
    return _AlarmScreenUNICOS_menuDisplayAlarmDetails(mSource, mAlarmData);
  }
  else if(iAnswer == (UN_NUM_AESVIEW + 300))
  {
    return _AlarmScreenUNICOS_menuInsertComment(mSource, mAlarmData);
  }
  return false;  // Not alarm specific menu action
}

/**
 * Copy content of active table row to clipboard
 * @param mSource Parameters of event in the table
 * @param mAlarmData All properties of selected alarm(s) in table
 * @return Always true
 */
private bool _AlarmScreenUNICOS_menuLineToClipboard(const mapping &mSource, const mapping &mAlarmData)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "()");
  string str = this.getTableRowAsString(mSource["row"]);
  setClipboardText(str);
  return true;
}

/**
 * Make visible all table columns
 * @param mSource Parameters of event in the table
 * @return Always true
 */
private bool _AlarmScreenUNICOS_menuShowAllColumns(const mapping &mSource)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "()");
  this.showAllTableColumns();  // The call is originated from AS EWO
  return true;
}

/**
 * Switch visibility of single column in table
 * @param mSource Parameters of event in the table
 * @param iColumnIdx Table column index
 * @return Always true
 */
private bool _AlarmScreenUNICOS_menuShowHideColumn(const mapping &mSource, int iColumnIdx)
{
  DebugFTN(ASNG_UN_DBG, __FUNCTION__ + "(" + iColumnIdx + ")");
  this.toggleTableColumnVisibility(iColumnIdx - 1);  // TODO: WinCC OA indices start with 1, C++ indices start with 0. Adjust somehow?
  return true;
}

/**
 * Display details of alarm in selected row
 * @param mSource Parameters of event in the table
 * @param mAlarmData All properties of selected alarm(s) in table
 * @return Always true
 */
private bool _AlarmScreenUNICOS_menuDisplayAlarmDetails(const mapping &mSource, const mapping &mAlarmData)
{
  string sDpId = mAlarmData[ALARM_PROP_FULL_DPE];
  int iDpeType = dpElementType(sDpId);

  // Check alarmtype - if sumalert / display sumalertdetails instead of details !
  if(isDpTypeSumAlert(iDpeType))
  {
    aes_displaySumalertDetails(sDpId, iRow, sDpProp);  // TODO ?
  }
  else
  {
    string sPanel = "vision/unAlarmScreenNg/AsNgAlarmDetailsUNICOS.pnl";
    // Start child panel for detail information
    string sTime = mAlarmData[ALARM_PROP_TIME];  // string in default format
    unGraphicalFrame_ChildPanelOnCentralModal(
          sPanel,
          "",
          makeDynString(
              "$dpid:" + dpSubStr(sDpId, DPSUB_SYS_DP_EL),
              "$time:" + sTime,
              "$count:" + mAlarmData[ALARM_PROP_INDEX]
            )
        );
  }
  return true;
}

/**
 * Set comment for alarm in selected row
 * @param mSource Parameters of event in the table
 * @param mAlarmData All properties of selected alarm(s) in table
 * @return Always true
 */
private bool _AlarmScreenUNICOS_menuInsertComment(const mapping &mSource, const mapping &mAlarmData)
{
  dyn_string exceptionInfo;
  AlarmScreenNgDisplay_showCommentPanel(mAlarmData, exceptionInfo);
  if(dynlen(exceptionInfo) > 0)
  {
    fwExceptionHandling_display(exceptionInfo);
  }
  return true;
}

/**
 * Acknowledge (multiple) alarms, selected in AS EWO
 * @param dmAlarms List of alarms data for alarms to be acknowledged
 */
private void _AlarmScreenUNICOS_ackMultiple(dyn_mapping &dmAlarms)
{
  AlarmScreenNg_checkForPartnersToAck(dmAlarms);
  AsNgAcknowledgeUNICOS acker;  // The class who will performa real acknowledgement
  acker.analyze(dmAlarms);

  string sMessage = acker.getDescription();
  if(acker.canAcknowledge()) {
    if(sMessage != "") {
      dyn_string dsReturn;
      dyn_float dfReturn;
      ChildPanelOnCentralModalReturn(
          "vision/MessageWarning2",
          getCatStr("sc", "Attention"),
          makeDynString("$1:" + sMessage,
                        "$2:" + getCatStr("general", "yes"),
                        "$3:" + getCatStr("sc", "no")),
          dfReturn, dsReturn);
      bool bContinueAck = false;
      if(dynlen(dfReturn) > 0) {
        if(dfReturn[1] == 1) {
          bContinueAck = true;
        }
      }
      if(!bContinueAck) {
        return;
      }
    }
  }
  else {
    if(sMessage != "") {
      ChildPanelOnCentralModal(
          "vision/MessageWarning",
          getCatStr("sc", "attention"),
          makeDynString("$1:" + sMessage));
    }
    return;
  }
  acker.acknowledge();
  ChildPanelOnCentralModal(
      "vision/MessageWarning",
      getCatStr("sc", "attention"),
      makeDynString("$1:Alarm acknowledgement completed"));
}

//@}

/** @name The replacement (copy) of some functions used from unAlarm componebt */
//@{
/**
  The replacement of function _unAlarmScreen_defaultPanel().<br>
  Action for click on "S" column of the alarm screen.

  @param  sDeviceName   string input, The DPE of the clicked device.
  @param  sDescription  string input, The description of the clicked device.

  @param Usage:
  Internal.
*/
private void _AlarmScreenUNICOS_defaultPanel(string sDeviceName, string sDescription)
{
  dyn_string dsParameters;
  dyn_string dsExceptions;


  unGenericDpFunctions_getParameters(sDeviceName, dsParameters);
  string sPath = dsParameters[UN_PARAMETER_PANEL];
  if(getPath(PANELS_REL_PATH, sPath) == "" || sPath=="")
  {
    bool bBad = true;
    if(dpExists(sPath))
    {
      // Check if of type unTreeWidget_treeNodeDPT
      if(dpTypeName(sPath) == unTreeWidget_treeNodeDPT)
      {
        string sPanelDp;
        string sDpType;
        string sPanelFile;
        dyn_string dsUserData;

        fwTree_getNodeUserData(_fwTree_getNodeName(sPath), dsUserData, dsExceptions);
        fwTree_getNodeDevice(_fwTree_getNodeName(sPath), sPanelDp, sDpType, dsExceptions);
        sPanelFile = unPanel_convertDpPanelNameToPanelFileName(sPanelDp);

        int iUserDataLength = dynlen(dsUserData);
        if(iUserDataLength <= 0)
        {
          // Case no userData
          dsUserData[1] = "";
          iUserDataLength = dynlen(dsUserData);
        }

        if((getPath(PANELS_REL_PATH, sPanelFile) != "") && (iUserDataLength > 0) && (sDpType == UN_PANEL_DPTYPE) && dpExists(sPanelDp))
        {
          string sPanelButton;
          unGenericDpFunctions_getContextPanel(sPanelDp, sPanelButton, dsExceptions);
          bBad = false;

          if(iUserDataLength > c_unPanelConfiguration_maxPanelNavigation)
          {
            // 11, give the sDpNodeName
            unGraphicalFrame_showUserPanel(sPanelFile, dsUserData[1], sPath);
            unGraphicalFrame_showContextualPanel(sPanelButton, sPath);
          }
          else
          {
            unGraphicalFrame_showUserPanel(sPanelFile, dsUserData[1], sPanelDp);
            unGraphicalFrame_showContextualPanel(sPanelButton, sPanelDp);
          }
        }
      }
    }

    if(bBad)
    {
      unGraphicalFrame_ChildPanelOnCentralModal("vision/MessageWarning","Warning",makeDynString("$1:" + unGenericDpFunctions_getAlias(sDeviceName) + getCatStr("unDB","EMPTYSYNOPTIC")));
    }
  }
  else
  {
    string sDpPanel = unPanel_convertPanelFileNameToDp(sPath);

    if(!dpExists(sDpPanel))
    {
      unPanel_create(sPath, sDpPanel, dsExceptions);
    }

    if (dynlen(dsExceptions) > 0)
    {
      unGraphicalFrame_ChildPanelOnCentral(sPath, unGenericDpFunctions_getAlias(sDeviceName) + " " + sDescription + " Synoptic", makeDynString());
    }
    else
    {
      unGraphicalFrame_showUserPanel(sPath, "", sDpPanel);

      string sPanelName;
      unGenericDpFunctions_getContextPanel(sDpPanel, sPanelName, dsExceptions);

      if (dynlen(dsExceptions) <= 0)
      {
        unGraphicalFrame_showContextualPanel(sPanelName, sDpPanel);
      }
    }
  }
}

//@}
